#include<libconfig.h>
#include<stdio.h>

//~ #define DEBUG

#include"mnb-configuration.h"
#include"mnb-utils.h"


//names used within the .conf files

#define TESTS "tests"

#define SIZE_CONFS "size_configurations"
#define NODE_CONFS "node_configurations"
#define TEST_COLLECTIONS "test_collections"
#define ARANGEMENT "test"


#define CONF "conf"
#define SIZE_CONF "sizeconf"
#define MNB_MIN_REPEATS "min_repeats"
#define MNB_MAX_DURATION "max_duration"


//~ static int  MNB_StringtoIntegerInSeconds(char * max_duration);
static char * MNB_GetNameFromSetting(config_setting_t * setting);
static char * MNB_GetStringSetting(config_setting_t * setting);
static MNB_Setting * MNB_GetChildSettingByName(char * name, MNB_Setting * parent);
static int MNB_FillSetting(MNB_Setting * typed_setting, config_setting_t * setting);
static int MNB_FillConfigurationFromLibconfig(config_t* config, MNB_Configuration * conf_to_fill);
static void MNB_ResolvTestSettingReferences(MNB_Setting * testsSetting, MNB_Setting * allSettings);
static struct MNB_List * MNB_ConvertToListOfTests(MNB_Setting * testsSetting, MNB_Configuration * conf);
static struct MNB_List * MNB_ConvertToTest(MNB_Setting * testSetting , MNB_Configuration * conf);
static void MNB_ExpandTestListForTests(MNB_Test * testTemplate, MNB_Setting * testsSetting, struct MNB_List * newTestList);
static void MNB_ExpandTestListForNodeConfs(MNB_Test * testTemplate, MNB_Setting * nodeConfSetting, struct MNB_List * newTestList);




/**
 * tries to retrieve the name of an libconfig setting
 * and return a pointer to a copy of the name
 * return NULL if the setting has no name
 * 
 * 
 **/
char * MNB_GetNameFromSetting(config_setting_t * setting){
	MNB_println("MNB_GetNameFromSetting");
	char	* constName = config_setting_name(setting);
	if(constName != NULL){
		char * newName = malloc(strlen(constName + 1));	
		strcpy(newName, constName);

		MNB_println(newName);
		return newName;
	}
	else
	{
		return constName;
	}
}



/**
 * returns the copy of the string value of a libconfig setting
 * 
 **/
char * MNB_GetStringSetting(config_setting_t * setting){
	const char * const_string = config_setting_get_string(setting);
	char * new_string = malloc(strlen(const_string + 1));
	strcpy(new_string, const_string);
	return new_string;
}

/**
 * Frees the momory used up by a MNB_Configuration
 * 
 **/
void MNB_FreeConfiguration(MNB_Configuration * conf){
	
	struct MNB_ListElement * iterator = conf->tests->first;
	while(iterator != NULL)
	{
		MNB_RemoveFromList(iterator->data, conf->tests);
		iterator = iterator->next;
		
	}
	MNB_DeleteList(conf->tests);
	free(conf);
}



/**
 * allocates the memory and sets default values for en MNB_Configuration
 * 
 **/
MNB_Configuration * MNB_CreateConfiguration(){
	MNB_println("CreateConfiguration");
	MNB_Configuration * config = malloc(sizeof(MNB_Configuration));
	config->tests = MNB_CreateList();
	config->minRepeats = MNB_DEFAULT_MIN_TEST_REPEATS;
	config->maxDuration = MNB_DEFAULT_MAX_TEST_DURATION;
	return config;
}



//======================================================================
/**
 * uses libconfig to read the information from a file and fill the 
 * provided struct with this information.
 * 
 **/
int MNB_ReadConfigurationFromFile(MNB_Configuration * conf, char * filename){
	int return_value = 0;
	MNB_println("FillConfigurationFromFile");

	//some libconfig initializations
	config_t * tmp_config = malloc(sizeof(config_t));
	config_init(tmp_config);

	//if we can't open the configuration file, than theres nothing to execute.
	conf->conf_file = fopen(filename, "r");
	if(conf->conf_file == NULL){
		//mpi_bench_perror_and_quit("couldn't open Configuration file");
		return_value  = -1;
		}
	
	config_read(tmp_config, conf->conf_file);
	
	/**
	 * 
	 * the following Block requires elements described in documentation for 
	 * libconfig v1.4.6
	 * however the standart ubuntu sources provide only v1.3.2 
	int config_read_success = config_read(tmp_config, conf->conf_file);
	if(config_read_success != CONFIG_TRUE){
		config_error_t error_type = config_error_type (tmp_config);
		switch(error_type){
			case CONFIG_ERR_NONE:
				mpi_bench_perror_and_quit("file wasn't parsed correctly but libconfig isn't admitting it");
			break;
			
			case CONFIG_ERR_FILE_IO:
				mpi_bench_perror_and_quit("couldn't read Configuration file");
			break;
			
			case CONFIG_ERR_PARSE:
				mpi_bench_perror_and_quit("couldn't parse Configuration file");
			break;
			
			default:
				mpi_bench_perror_and_quit("file wasn't parsed correctly, no idea what went wrong, you propably should check your .cfg file");
			break;
				
			}
		}
	**/

	return_value = MNB_FillConfigurationFromLibconfig(tmp_config, conf);


	//the libconfig config isn't requered anymode, so we are deleting it
	config_destroy(tmp_config);
	free(tmp_config);
	
	
	return return_value;
}



/**
 * tries to retrieve a subsetting of parent with the given name and return it
 * 
 * returns NULL if no such setting can be found
 * 
 **/
MNB_Setting * MNB_GetChildSettingByName(char * name, MNB_Setting * parent){
	MNB_printf("MNB_GetChildSettingByName:%s\n", name );
	if(parent->type == TYPE_LIST){

		struct MNB_ListElement * iterator = ((struct MNB_List *)parent->data)->first;
		MNB_printf("setting size:%d\n",((struct MNB_List *)parent->data)->size);
		if(iterator == NULL){
			MNB_printf("setting name:%s\n", "iterator == NULL");
		}
		
		while(iterator != NULL) {
			MNB_Setting * childSetting = (MNB_Setting *)iterator->data;
			MNB_printf("setting name:%s\n", childSetting->name);
			if(childSetting->name != NULL){
				MNB_println(childSetting->name);
				if(strcmp(childSetting->name, name) == 0){
					MNB_println("subSetting found");
					return iterator->data;
				}
				iterator = iterator->next;
			}
		}

	}
	//else 
	return NULL;
	
	
	
}








/**
 * returns the childsetting with given name
 * (only the first level of subsettings is considered)
 * 
 * NULL if setting could not be found
 **/
#define MNB_GetChildByName(setting, name) config_setting_get_member(setting, name)


/**
 * converts the libconfig setting to a MNB_Setting 
 * 
 * uses recursive calls for list and array settings
 * returns -1 if for an NULL pointer at the libconfig setting
 * or an unexpected setting type
 * 
 **/
int MNB_FillSetting(MNB_Setting * mnbSetting, config_setting_t * setting){
	int return_value = 0;
	
	MNB_println("fillSetting");
	if(setting == NULL){
		return -1;
	}
		
	//getting some setting type independent information about the setting
	int 	settingType = config_setting_type(setting),
			settingLength = config_setting_length(setting);
	char * settingName = MNB_GetNameFromSetting(setting);
	mnbSetting->name = settingName;
			
	MNB_println("swith on type");
		
		
	//and now we are doint the things that are differen for each type of setting
	//which is to set the type and the value
	switch(settingType){
		case CONFIG_TYPE_STRING:
			MNB_println("its a string setting");
			//our get MNB_GetStringSetting funktion already provides us
			//with a copy so we don't have to do it here 	
			char * confString = MNB_GetStringSetting(setting);
			
			mnbSetting->type = TYPE_STRING;
			mnbSetting->data = confString;
			
			MNB_printf("setting:%s\n",confString);
		
			break;
			
		case CONFIG_TYPE_INT:
			MNB_println("its a int setting");
			//We have no getIntFunktion, so we have to create a copy here
			int confInt = config_setting_get_int(setting);
			int * newInt = malloc(sizeof(int));
				* newInt = confInt;
				
			mnbSetting->type = TYPE_INT;
			mnbSetting->data = newInt;
			
				MNB_printf("setting:%d\n",confInt);
			break;
			
			//all three types will be represented as list settings
		case CONFIG_TYPE_LIST:
		case CONFIG_TYPE_GROUP:
		case CONFIG_TYPE_ARRAY:
			MNB_println("its a list setting");
			//create a new list
			struct MNB_List * subConfList = MNB_CreateList();;
			
			//setting everything we know about the setting
			mnbSetting->type = TYPE_LIST;
			mnbSetting->data = subConfList;
			
			
			//filling it with subsettings
			config_setting_t * subSetting;
			int j; 
			for(j = 0; j < settingLength; j = j + 1)
				{
					subSetting =  config_setting_get_elem(setting, j);
					MNB_Setting * newSubSetting = MNB_CreateSetting();
					MNB_FillSetting(newSubSetting, subSetting);
					MNB_AddToList((void *) newSubSetting, subConfList);
					
				}
			
				MNB_printf("struct MNB_List name:%s\n", subConfList->name);
				MNB_printf("struct MNB_List name:%s\n", ((LList *)mnbSetting->data)->name);
			
			break;
			
			default:
				MNB_println("unexpected settig type");
				return_value = -1;
			
			break;
		
	}
		
		
	
	

	return return_value;
}





/**
 * fills mpi-bench Configuration struct with information from 
 * libconfigs
 * 
 * 
 * convert file to list of MNB_Settings
 * get TestList.
 * Resolv settings in test list
 * convert to list of MNB_Tests
 * 
 **/
int MNB_FillConfigurationFromLibconfig(config_t* config, MNB_Configuration * confToFill){
	int return_value = 0;

	MNB_println("MNB_FillConfigurationFromLibconfig");

	config_setting_t * rootSetting = config_root_setting (config);

	//checking if some boundaries for tests are provided. Using defaults
	//otherwise
	
	MNB_Setting * allSettings = MNB_CreateSetting();
	allSettings->type = TYPE_LIST;
	
	
	//converting the libconfig-settings from the file to MNB_Setting(lists)
	MNB_FillSetting(allSettings, rootSetting);
	
	
	
	//getting the tests
	MNB_Setting * testsSetting = MNB_GetChildSettingByName(TESTS, allSettings);
	//without the tests setting there is nothing to execute
	if(testsSetting == NULL){
		MNB_Abort("Could't find tests setting. Reason might be a missing semicolomn in the configuration file");
	}
	
	
	
	//replacing the references to preferences defined in node_conf, sizes, etc
	MNB_ResolvTestSettingReferences(testsSetting, allSettings);
	
	#ifdef DEBUG
		MNB_PrintSetting(allSettings);
	#endif
	
	confToFill->tests = MNB_ConvertToListOfTests(testsSetting, confToFill);

return return_value;
}  


/**
 * Goes throut the three of test setting and tryes to replace all the
 * references throu the settings they are pointing at
 * 
 * 
 **/
void MNB_ResolvTestSettingReferences(MNB_Setting * testsSetting, MNB_Setting * allSettings){
	//MNB_println("resolving the references"); <- becomes to annoying, because of the recursive calls
	//we have to go through all the levels of the tests Setting
	if(testsSetting->type == TYPE_LIST){
		struct MNB_ListElement * iterator = ((struct MNB_List *)testsSetting->data)->first;
		while(iterator != NULL){
			
			MNB_ResolvTestSettingReferences((MNB_Setting *)iterator->data, allSettings);
			iterator = iterator->next;
		}
	}
	//if testsSetting isn't a list anymore, we are at the lowest level
	//we need to check if the strings here are references to node_confs, sizes, etc
	else if(testsSetting->type == TYPE_STRING)//integers can't be references
	{
		//if the data is a reference to this setting..  we can't rely on the setting to have a name, so we check it before using
		if((allSettings->name != NULL) && (strcmp(testsSetting->data, allSettings->name) == 0)){
				//~ printf("try to resolf:%s\n", testsSetting->name);
				testsSetting->type = allSettings->type;
				free(testsSetting->data);
				testsSetting->data = allSettings->data;	
		
		} else {//if the reference was not found, lets look at the subsettings of allSettings
			 
			if((allSettings)->type == TYPE_LIST){
				struct MNB_ListElement * iterator = ((struct MNB_List *)allSettings->data)->first;
				while(iterator != NULL){
					MNB_ResolvTestSettingReferences(testsSetting, (MNB_Setting *)iterator->data);
					iterator = iterator->next;
				}
			}
		}
	}
}
/**
 * We have to convert each element of the tests setting to a list of MNB_Tests
 * 
 **/
struct MNB_List * MNB_ConvertToListOfTests(MNB_Setting * testsSetting, MNB_Configuration * conf){
	MNB_println("converting to tests");

	struct MNB_List * testsList = MNB_CreateList();
	
	if(testsSetting->type == TYPE_LIST){
		struct MNB_ListElement * iterator = ((struct MNB_List *)testsSetting->data)->first;
		while(iterator != NULL){
			struct MNB_List * newTestList = MNB_ConvertToTest((MNB_Setting *)iterator->data, conf);
			testsList = MNB_ConcatList(testsList, newTestList);
			iterator = iterator->next;
		}
		MNB_printf("all tests created. %d tests in list\n", testsList->NumEl);
		return testsList;
	} else {
		return NULL;
	}
	
	
}


struct MNB_List * MNB_ConvertToTest(MNB_Setting * testSetting, MNB_Configuration * conf){
	MNB_printf("MNB_ConvertToTest:%s\n", testSetting->name);
	//wir setzten minRepeats, maxDuration und sizes
	//wenn tests eine liste ist, expand ansonsten einfach zuweisen
	//wenn conf eine liste von listen ist expand ansonsten einfach zuweisen
	
	
	
	//~ struct MNB_List * testList = MNB_CreateList();
	MNB_Test * test = MNB_CreateTest();
	struct MNB_List * testList = MNB_CreateList();
	
	//set minRepeats
	MNB_Setting * minRepeatsSetting = MNB_GetChildSettingByName(MNB_MIN_REPEATS, testSetting);
	if(minRepeatsSetting != NULL){
		test->minRepeats = *((int *)minRepeatsSetting->data); 
	} else {
		test->minRepeats = conf->minRepeats; 
	}
	
	//set maxRepeats
	MNB_Setting * maxDurationSetting = MNB_GetChildSettingByName(MNB_MAX_DURATION, testSetting);
	if(maxDurationSetting != NULL){
		test->maxDuration = *((int *)maxDurationSetting->data); 
	} else {
		test->maxDuration = conf->maxDuration; 
	}
	
	//set sizes
	MNB_Setting * sizesSetting = MNB_GetChildSettingByName(SIZE_CONF, testSetting);
	if(sizesSetting != NULL){
		test->sizes = sizesSetting->data; 
	} else {
		printf("test without sizes is pointless. something went propably wrong.\n you schould check your conf file");
		//todo QUIT?
		test->sizes = NULL; 
	}
	
	//so this was the simple part.
	//there might be several tests and nodeconfs for one test in the conf file
	//we have to create a copy of the test sofar for each
	
	
	
	MNB_Setting * testsSetting = MNB_GetChildSettingByName(TESTS, testSetting);
	MNB_ExpandTestListForTests(test, testsSetting, testList);
	
	
	//this list is goint to become the list of complete tests, 
	//we are goint to return it
	struct MNB_List * completeList = MNB_CreateList();
	
	
	MNB_Setting * nodesSetting = MNB_GetChildSettingByName(CONF, testSetting);
	
	struct MNB_ListElement * iterator = testList->first;
	while(iterator != NULL){
		//we have to assume we are having a list of tests by now and we
		//have to expand each one of them
		MNB_Test * iteratorTest = (MNB_Test *)iterator->data;
		MNB_ExpandTestListForNodeConfs(iteratorTest, nodesSetting, completeList);
		
		iterator = iterator->next;
	}
	
	MNB_printf("test list conversion complete. %d tests created\n", completeList->NumEl);
	return completeList;
}


/**
 * for each element in the tests Setting a copy of the testTemplate is created 
 * and the element of tests Setting is set for it as the tests
 * 
 * 
 **/
void MNB_ExpandTestListForTests(MNB_Test * testTemplate, MNB_Setting * testsSetting, struct MNB_List * newTestList){
	
	switch(testsSetting->type){
		case TYPE_STRING:;
			MNB_Test * newTest = MNB_CopyTest(testTemplate);
			MNB_Setting * newSetting = MNB_CreateSetting();
			newSetting->type = TYPE_STRING;
			newSetting->data = testsSetting->data;
			newTest->test = newSetting->data ;
			MNB_AddToList(newTest, newTestList);
		break;
		
		case TYPE_LIST:;
			struct MNB_ListElement * iterator = ((struct MNB_List *)testsSetting->data)->first;
			while(iterator != NULL){
				MNB_Setting * iteratorSetting = (MNB_Setting *)iterator->data;
				MNB_ExpandTestListForTests(testTemplate, iteratorSetting, newTestList);
				iterator = iterator->next;
			}
		break;
		
		default:
		;
		//TODO: mit einem fehler abbrechen
		break;
	}
}



/**
 * for each node configuration in nodeConfSetting, a copy of the testTemplate
 * is created and the current configuration is assigned to it as the nodeConf
 * 
 **/
void MNB_ExpandTestListForNodeConfs(MNB_Test * testTemplate, MNB_Setting * nodeConfSetting, struct MNB_List * newTestList){
	MNB_println("MNB_ExpandTestListForNodeConfs");
	struct MNB_List * nodeConfList = (struct MNB_List *)nodeConfSetting->data;
	MNB_Setting * firstNodeConfSetting = (MNB_Setting *)(nodeConfList->first->data);
	//the separation has to happen based on the type of the first element of the
	//nodeConfList, since all node conf settings are lists, we have to figure out
	//if the first element is a list, meaning its a list of several node conf settings
	//of it its an integer, meaning its a simple node conf
	
	switch(firstNodeConfSetting->type)
	{
		case TYPE_STRING:
		case TYPE_INT:;
			MNB_Test * newTest = MNB_CopyTest(testTemplate);
			newTest->nodeConf = nodeConfSetting->data;
			MNB_AddToList(newTest, newTestList);
		
		break;
		case TYPE_LIST:;
			
			struct MNB_ListElement * iterator =  nodeConfList->first;
			while(iterator != NULL)
			{
				MNB_Setting * iteratorSetting = (MNB_Setting *)iterator->data;
				MNB_ExpandTestListForNodeConfs( testTemplate, iteratorSetting, newTestList);
				iterator = iterator->next;
			}
		break;
		
		default:
		//TODO mit fehlermeldung abbrechen
		break;
	}
}

