//~ #define DEBUG

#include"mnb-scheduling.h"


static int MNB_CheckIfSchedulablePerfect(MNB_Test * test, MNB_Environment * env, struct MNB_List * nodeListForTest);
static int MNB_CheckIfSchedulableGreedy(MNB_Test * test, MNB_Environment * env, struct MNB_List * nodeListForTest);
/**
 * Creates a copy of the given nodeList, containing only thouse
 * with MNB_NODE_FREE status
 **/
static struct MNB_List * MNB_CopyFreeNodesList(struct MNB_List * nodeList);



/**
 * Sorts the list of tests by their node and rank requerements
 * 
 * works with an quicksort/devide and conquer approch
 **/
struct MNB_List * MNB_SortTestList(struct MNB_List * testList){
	//the list is so small, we have nothing else to do
	if(testList->size <= 1){
		return testList;
		
		
	//for bigger lists, we have the usual thing to do 
	} else {  
		//choose an element
		int splitElementOffset = testList->size / 2;
		MNB_Test * splitElement = NULL;
		struct MNB_ListElement * iterator = testList->first;
		while(iterator != NULL){
			MNB_Test * test = iterator->data;
			if(splitElementOffset == 0){
				splitElement = test;
				MNB_RemoveFromList(test, testList);
				break;
			} 
			splitElementOffset = splitElementOffset - 1;
			iterator = iterator->next;
		}
		
		
		struct MNB_List * smallerList, * biggerList;
		smallerList = MNB_CreateList();
		biggerList = MNB_CreateList();
		
		//compare all other elements to the chosen and devide them in a smaller and a bigger list
		iterator = testList->first;
		while(iterator != NULL){
			MNB_Test * test = (MNB_Test *)iterator->data;
			if(MNB_GetNodeRequerment(test) > MNB_GetNodeRequerment(splitElement)){
				MNB_AddToList(test, biggerList);
			} else {
				if(MNB_GetNodeRequerment(test) == MNB_GetNodeRequerment(splitElement)) {
					//if the host requerement is the same, the proc requerement has to decide
					if(MNB_GetRankRequerment(test) >= MNB_GetRankRequerment(splitElement)){
						MNB_AddToList(test, biggerList);
					} else {
						MNB_AddToList(test, smallerList);
					}
				}
				//if the host requerement isn't bigger or equal, it must be smaller
				else {		
					MNB_AddToList(test, smallerList);
				}
			}
			iterator = iterator->next;
		}
		
		
		//sort the two lists
		struct MNB_List 	* sortedBiggerList  = MNB_SortTestList(biggerList),
				* sortedSmallerList = MNB_SortTestList(smallerList);
		
		//and finaly build everything back together
				
		//since we are inserting Elements in the end
		MNB_AddToList(splitElement, sortedBiggerList);

		return MNB_ConcatList(sortedBiggerList, sortedSmallerList);
	
	}
}

/**
 * Sorts the list of nodes by the amoount of nodes running on them
 * 
 * works with an quicksort/devide and conquer approch
 **/

struct MNB_List * MNB_SortNodeList(struct MNB_List * nodeList){
	MNB_println("MNB_SortNodeList");
	//the list is so small, we have nothing else to do
	if(nodeList->size <= 1){
		return nodeList;
		
		
	//for bigger lists, we have the usual thing to do 
	} else {  
		//choose an element, somewhere in the middle
		int splitElementOffset = nodeList->size / 2;
		MNB_Node * splitElement = NULL;
		struct MNB_ListElement * iterator = nodeList->first;
		while(iterator != NULL){
			MNB_Node * node = iterator->data;
			if(splitElementOffset == 0){
				splitElement = node;
				MNB_RemoveFromList(node, nodeList);
				break;
			} 
			splitElementOffset = splitElementOffset - 1;
			iterator = iterator->next;
		}
		
		
		struct MNB_List * smallerList, * biggerList;
		smallerList = MNB_CreateList();
		biggerList = MNB_CreateList();
		
		//compare all other elements to the chosen and devide them in a smaller and a bigger list
		iterator = nodeList->first;
		while(iterator != NULL){
			MNB_Node * node = (MNB_Node *)iterator->data;
			if(node->ranks->size >= splitElement->ranks->size){
				MNB_AddToList(node, biggerList);
			} else {
				//less nodes -> smaller List
				MNB_AddToList(node, smallerList);
			}
			iterator = iterator->next;
		}
		
		
		//sort the two lists
		struct MNB_List 	* sortedBiggerList  = MNB_SortNodeList(biggerList),
				* sortedSmallerList = MNB_SortNodeList(smallerList);
		
		//and finaly build everything back together
				
		//since we are inserting Elements in the end
		MNB_AddToList(splitElement, sortedBiggerList);

		
		return MNB_ConcatList(sortedBiggerList, sortedSmallerList);
	
	}
}



/**
 * looks for the first schedulable test and returns it.
 * if none can be scheduled NULL is returned
 * 
 * 
 **/
MNB_Test * MNB_GetFirstSchedulableTest(struct MNB_List * tests, MNB_Environment * env, struct MNB_List * nodeList){
	MNB_println("MNB_GetFirstSchedulableTest");
	#ifdef DEBUG
	
	if(tests == NULL){
		printf("%s\n", "tests is an empty pointer");
	} else { if(tests->size == 0){
		printf("%s\n", "tests list is empty");
	}}
	
	#endif
	
	
	struct MNB_ListElement * iterator = tests->first;
	while(iterator != NULL){
		MNB_Test * test = (MNB_Test *)iterator->data;
		if(MNB_CheckIfSchedulablePerfect(test, env, nodeList)){
			MNB_println("perfect schedulable test found");
			
			//marking the nodes as busy so we don't try to use them for another test
			MNB_SetNodesStatus(nodeList, MNB_NODE_BUSY);
			return test;
		}
		iterator = iterator->next;
	}
	//if no perfect match for a test could be found, lets try the same greedy
	iterator = tests->first;
	while(iterator != NULL){
		MNB_Test * test = (MNB_Test *)iterator->data;
		if(MNB_CheckIfSchedulableGreedy(test, env, nodeList)){
			MNB_println("greedy schedulable test found");
			
			//marking the nodes as busy so we don't try to use them for another test
			MNB_SetNodesStatus(nodeList, MNB_NODE_BUSY);
			return test;
		}
		iterator = iterator->next;
	}
	
	
	MNB_println("no schedulable test found");
	return NULL;
}


/**
 * checks if there are enough free Hosts with free ranks available in
 * the environment to schedule the test
 * 
 * 
 * we are using a depth first search approach
 **/
int MNB_CheckIfSchedulablePerfect(MNB_Test * test, MNB_Environment * env, struct MNB_List * nodeListForTest){
	MNB_printf("MNB_CheckIfSchedulablePerfect requerements host:%d proc:%d\n", 
											MNB_GetNodeRequerment(test),
											MNB_GetRankRequerment(test));
	//we can't use the busy Nodes anyway
	struct MNB_List * freeNodesList = MNB_CopyFreeNodesList(env->testerNodes);

	
	
	//cince we aren't really making recursive calls, freeNodesList helps to emulate the call
	struct MNB_ListElement * freeNodesListCurrentFirst = freeNodesList->first;
	//since we aren't working recursively, we don't have a real return value
	int pseudoReturnValue = TRUE;
	
	
	int nodesInList = 0, ranksInList = 0;
	while(	nodesInList != MNB_GetNodeRequerment(test) || //as long as our requerements aren't met
			ranksInList != MNB_GetRankRequerment(test)
			){
	
	MNB_printf("Node requerment:%d ,rank requerment:%d, nodes in list:%d, ranks in list:%d\n",
				MNB_GetNodeRequerment(test),
				MNB_GetRankRequerment(test),
				nodesInList,
				ranksInList);
		
		//the simple case, we can just add another node
		if(	(nodesInList < MNB_GetNodeRequerment(test)) &&
			(ranksInList < MNB_GetRankRequerment(test)) &&  
			(pseudoReturnValue == 1)){	//this line makes shure we aren't looking for candidates, 
										//if we allready did in the iteration before and didn't find any
										//0 is used because it won't compile with TRUE;
				MNB_println("we shoul try to add another node");
				int ranksWeStillNeed = MNB_GetRankRequerment(test) - ranksInList;
				int candidateFound = FALSE;
				
				//find a node that can be added
				struct MNB_ListElement * iterator = freeNodesListCurrentFirst;
				while(iterator != NULL){
					MNB_Node * nextNodeCandidate = (MNB_Node *)iterator->data;
					MNB_printf("ranks we need:%d , next candidate has %d\n",ranksWeStillNeed, nextNodeCandidate->ranks->size);
					if(nextNodeCandidate->ranks->size <= ranksWeStillNeed){ 
						MNB_println("candidate found");
						//and add it
						MNB_AddToList(nextNodeCandidate, nodeListForTest);
						nodesInList = nodesInList + 1;
						ranksInList = ranksInList + nextNodeCandidate->ranks->size;
						candidateFound = TRUE;
						freeNodesListCurrentFirst = iterator->next;
						break;
					}
					iterator = iterator->next;
				}
				
				if(!candidateFound){
					//getting here means, no candidate could be found to add
					MNB_println("no candidate found");
					pseudoReturnValue = FALSE;
				}
				continue;  
			
		} else {
			pseudoReturnValue = TRUE;
			//one way or another, if the requerements arent met, and we can't add nodes,
			//we have to get rid of some of the ones allready added
			
			//but if there is nothing left we could remove, then the test isn't schedulable
			if(nodesInList == 0){
				MNB_println("no valid test combination could be found");
				return FALSE;
			}
			
			MNB_Node * lastNodeAdded = nodeListForTest->last->data;
			
			MNB_RemoveFromList(lastNodeAdded, nodeListForTest);
			nodesInList = nodesInList - 1;
			ranksInList = ranksInList - lastNodeAdded->ranks->size;
			
			//and we place the current first pointer after the element we have just removed, so
			//we wan't try to add it again.
			struct MNB_ListElement * iterator = freeNodesList->first;
			while(iterator != NULL){
				if(iterator->data == lastNodeAdded){
					freeNodesListCurrentFirst = iterator->next;
					break;
				}
				iterator = iterator->next;
			}	
		
		}
	}
	//so we have found an configuration, and could leave the loop
	//this means this test is schedulable
	
	return TRUE;
	
}

/**
 * checks if there are enough free Hosts with free ranks available in
 * the environment to schedule the test
 * 
 * its allmost an exact copy of MNB_CheckIfSchedulablePerfect but with more 
 * loose requeremets
 * 
 * we are using a Greedy search approach, meaning that we will accept a
 * conbination containing the right amount of nodes, but more ranks than we need
 **/
int MNB_CheckIfSchedulableGreedy(MNB_Test * test, MNB_Environment * env, struct MNB_List * nodeListForTest){
	MNB_printf("MNB_CheckIfSchedulableGreedy requerements host:%d proc:%d\n", 
											MNB_GetNodeRequerment(test),
											MNB_GetRankRequerment(test));
	//we can't use the busy Nodes anyway
	struct MNB_List * freeNodesList = MNB_CopyFreeNodesList(env->testerNodes);

	
	
	//cince we aren't really making recursive calls, freeNodesList helps to emulate the call
	struct MNB_ListElement * freeNodesListCurrentFirst = freeNodesList->first;
	//since we aren't working recursively, we don't have a real return value
	int pseudoReturnValue = TRUE;
	
	
	int nodesInList = 0, ranksInList = 0;
	while(	nodesInList != MNB_GetNodeRequerment(test) || //as long as our requerements aren't met
			ranksInList < MNB_GetRankRequerment(test)	){ //we can accept more ranks than we actualy need, we don't have to use them all
	
	MNB_printf("Node requerment:%d ,rank requerment:%d, nodes in list:%d, ranks in list:%d\n",
				MNB_GetNodeRequerment(test),
				MNB_GetRankRequerment(test),
				nodesInList,
				ranksInList);
		
		//the simple case, we can just add another node
		if(	(nodesInList < MNB_GetNodeRequerment(test)) &&
			(pseudoReturnValue == 1)){	//this line makes shure we aren't looking for candidates, 
										//if we allready did in the iteration before and didn't find any
										//0 is used because it won't compile with TRUE;
				MNB_println("we shoul try to add another node");
				int candidateFound = FALSE;
				
				//find a node that can be added
				struct MNB_ListElement * iterator = freeNodesListCurrentFirst;
				while(iterator != NULL){
					MNB_Node * nextNodeCandidate = (MNB_Node *)iterator->data;
						MNB_println("candidate found");
						//and add it
						MNB_AddToList(nextNodeCandidate, nodeListForTest);
						nodesInList = nodesInList + 1;
						ranksInList = ranksInList + nextNodeCandidate->ranks->size;
						candidateFound = TRUE;
						freeNodesListCurrentFirst = iterator->next;
						break;
					iterator = iterator->next;
				}
				
				if(!candidateFound){
					//getting here means, no candidate could be found to add
					//meaning the nodes which are in the list right now make it impossible to find a compbination
					//so we have to set the pseudoReturnValue to FALSE, to remove an element in the next iteration
					MNB_println("no candidate found");
					pseudoReturnValue = FALSE;
				}
				continue;  
			
		} else {
			pseudoReturnValue = TRUE;
			//one way or another, if the requerements arent met, and we can't add nodes,
			//we have to get rid of some of the ones allready added
			
			//but if there is nothing left we could remove, then the test isn't schedulable
			if(nodesInList == 0){
				MNB_println("no valid test combination could be found");
				return FALSE;
			}
			
			MNB_Node * lastNodeAdded = nodeListForTest->last->data;
			
			MNB_RemoveFromList(lastNodeAdded, nodeListForTest);
			nodesInList = nodesInList - 1;
			ranksInList = ranksInList - lastNodeAdded->ranks->size;
			
			//and we place the current first pointer after the element we have just removed, so
			//we wan't try to add it again.
			struct MNB_ListElement * iterator = freeNodesList->first;
			while(iterator != NULL){
				if(iterator->data == lastNodeAdded){
					freeNodesListCurrentFirst = iterator->next;
					break;
				}
				iterator = iterator->next;
			}	
		
		}
	}
	//so we have found an configuration, and could leave the loop
	//this means this test is schedulable
	
	return TRUE;
	
}




/**
 * Creates a copy of the given nodeList, containing only thouse
 * with MNB_NODE_FREE status
 **/
struct MNB_List * MNB_CopyFreeNodesList(struct MNB_List * nodeList){
	MNB_println("MNB_CopyFreeNodesList");
	#ifdef DEBUG
	
	if(nodeList == NULL){
		printf("%s\n", "MNB_CopyFreeNodesList:nodeList is an empty pointer");
	} else { 
		printf("MNB_CopyFreeNodesList:nodeList->size: %d\n", nodeList->size);
	}
	
	#endif
	
	
	struct MNB_List * freeNodesList = MNB_CreateList();
	
	//goint through the list of nodes
	struct MNB_ListElement * iterator = nodeList->first;
	while(iterator != NULL){
		MNB_Node * node = (MNB_Node *)iterator->data;
		if(node->status == MNB_NODE_FREE){
			//adding the right ones to the list
			MNB_AddToList(node, freeNodesList);
		} else {
			MNB_println("busy node");
		}
		
		iterator = iterator->next;
	}
	MNB_printf("MNB_CopyFreeNodesList:created a list of %d free nodes\n", freeNodesList->size);
	return freeNodesList;
}




/**
 * sets the status of all nodes in the list to the given value
 * 
 * 
 **/
void MNB_SetNodesStatus(struct MNB_List * nodes, enum MNB_NodeStatus status){
	struct MNB_ListElement * iterator = nodes->first;
	while(iterator != NULL){
		MNB_Node * node = (MNB_Node *)iterator->data;
		node->status = status;
		iterator = iterator->next;
	}
}


