//~ #define DEBUG

#include"mnb-stdtests.h"


double reduceTo0(long size, MPI_Comm comm){
  size = size - (size % 8);
  
  double * sendbuf = mallocWithCheck(size);
  double * recvbuf = mallocWithCheck(size);
  
  MNB_println("reduceTo0 function");
  
  STARTTIME
  int ret = MPI_Reduce(sendbuf, recvbuf, size / 8 , MPI_DOUBLE, MPI_SUM, 0, comm);
  CHECKRET
  GETTIME
  

  free(sendbuf);
  free(recvbuf);
  
  return e;
}

double allreduce(long size, MPI_Comm comm){
  size = size - (size % 8);
  
  double * sendbuf = mallocWithCheck(size);
  double * recvbuf = mallocWithCheck(size);
  
  MNB_println("allreduce function");
  
  STARTTIME
  int ret = MPI_Allreduce(sendbuf, recvbuf, size / 8, MPI_DOUBLE, MPI_SUM, comm);
  CHECKRET
  GETTIME
	
  free(sendbuf);
  free(recvbuf);
  
  return e;
}


double bcast(long size, MPI_Comm comm){
	MNB_printf("%s\n", "bcast");
	size = size - (size % 8);
	double * buffer = mallocWithCheck(size);
  
	STARTTIME  
	int ret = MPI_Bcast(buffer, size / 8, MPI_DOUBLE, 0, comm);
	CHECKRET
	GETTIME
	free(buffer);
	
	return e;
}


double barrier(long unused, MPI_Comm comm){
	MNB_printf("%s\n", "barrier");
	STARTTIME
	MPI_Barrier(comm);
	double e = MPI_Wtime() - t;  

	return e;
}


double gather(long size, MPI_Comm comm){
	MNB_printf("%s\n", "gather");
	size = size - (size % 8);
	double * sendbuf = mallocWithCheck(size);
	int nproc;
	MPI_Comm_size(comm, &nproc);

	double * recvbuf = mallocWithCheck(size * nproc);

	STARTTIME
	int ret =     MPI_Gather(sendbuf, size / 8, MPI_DOUBLE, recvbuf, size/ 8, MPI_DOUBLE, 0, comm);
	CHECKRET
	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}

double allgather(long size, MPI_Comm comm){
	MNB_printf("%s\n", "allgather");
	size = size - (size % 8);
	double * sendbuf = mallocWithCheck(size);
	int nproc;
	MPI_Comm_size(comm, &nproc);

	double * recvbuf = mallocWithCheck(size * nproc);

	STARTTIME
	int ret =     MPI_Allgather(sendbuf, size / 8, MPI_DOUBLE, recvbuf, size/ 8, MPI_DOUBLE, comm);
	CHECKRET
	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}



double scatter(long size, MPI_Comm comm){
	MNB_printf("%s\n", "scatter");
	size = size - (size % 8);
	int nproc;
	MPI_Comm_size(comm, &nproc);
	double * sendbuf = mallocWithCheck(size * nproc);
	double * recvbuf = mallocWithCheck(size);

	STARTTIME
	int ret =     MPI_Scatter(sendbuf, size / 8, MPI_DOUBLE, recvbuf, size/ 8, MPI_DOUBLE, 0, comm);
	CHECKRET
	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}


double sendRecvRightNeighbour(long size, MPI_Comm comm){
	MNB_printf("%s\n", "sendRecvRightNeighbour");
	size = size - (size % 8);
	int nproc, rank;
	MPI_Comm_size(comm, &nproc);
	MPI_Comm_rank(comm, &rank);
	double * sendbuf = mallocWithCheck(size);
	double * recvbuf = mallocWithCheck(size);

	const int dest = rank + 1 == nproc ? 0 : rank + 1 ;
	const int source = rank - 1 >= 0 ? rank - 1 : nproc -1;

	STARTTIME
	int ret =       MPI_Sendrecv( sendbuf, size / 8, MPI_DOUBLE, dest, 4711,
	recvbuf, size / 8, MPI_DOUBLE, source, 4711, comm, MPI_STATUS_IGNORE);
	CHECKRET
	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}


double sendRecvPaired(long size, MPI_Comm comm){
	MNB_printf("%s\n", "sendRecvPaired");
	size = size - (size % 8);

	int nproc, rank;
	MPI_Comm_size(comm, &nproc);
	MPI_Comm_rank(comm, &rank);

	if(nproc % 2 == 1){
	/* test is Run only with equal number of processes */
		return 0;
	}

	double * sendbuf = mallocWithCheck(size);
	double * recvbuf = mallocWithCheck(size);


	const int dest = rank % 2 == 0 ? rank + 1 : rank -1;

	STARTTIME
	int ret =  MPI_Sendrecv( sendbuf, size / 8, MPI_DOUBLE, dest, 4711,
	recvbuf, size / 8, MPI_DOUBLE, dest, 4711, comm, MPI_STATUS_IGNORE);
	CHECKRET
	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}



double sendToRoot(long size, MPI_Comm comm){
	
	size = size - (size % 8);
	MNB_printf("sendToRoot size : %ld , size mod 8: %ld \n", size , size % 8);
	
	int nproc, rank;
	MPI_Comm_size(comm, &nproc);
	MPI_Comm_rank(comm, &rank);

	double * sendbuf = mallocWithCheck(size);
	double * recvbuf = mallocWithCheck(size);

	STARTTIME
	if(rank == 0){
		for(int i=1; i < nproc; i++){
			int ret = MPI_Recv( sendbuf, size / 8, MPI_DOUBLE, i, 4711, comm, MPI_STATUS_IGNORE);
			CHECKRET
		}
	}else{
		int ret = MPI_Send( sendbuf, size / 8, MPI_DOUBLE, 0, 4711, comm);
		CHECKRET
	}

	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}

double sendRecvToRoot(long size, MPI_Comm comm){
	MNB_printf("%s\n", "sendRecvToRoot");
	
	size = size - (size % 8);
	  
	int nproc, rank;
	MPI_Comm_size(comm, &nproc);
	MPI_Comm_rank(comm, &rank);
	  
	double * sendbuf = mallocWithCheck(size);
	double * recvbuf = mallocWithCheck(size);

	STARTTIME
	if(rank == 0){
		for(int i=1; i < nproc; i++){
			int ret = MPI_Sendrecv( sendbuf, size / 8, MPI_DOUBLE, i, 4711, recvbuf, size/ 8, MPI_DOUBLE, i, 4711, comm, MPI_STATUS_IGNORE);
			CHECKRET
		}
	}else{
		int ret = MPI_Sendrecv( sendbuf, size / 8, MPI_DOUBLE, 0, 4711, recvbuf,  size/ 8, MPI_DOUBLE, 0, 4711, comm, MPI_STATUS_IGNORE);
		CHECKRET
	}

	GETTIME
	free(sendbuf);
	free(recvbuf);

	return e;
}



/**
 * tries to find a function with the given name and return a pointer to it
 * returns NULL if function can't be found
 * 
 * 
 **/
testFunc MNB_GetTestByName(char * test_name, char * source)
{
	if(strcmp(test_name, "reduceTo0") == 0)
	{
		return  reduceTo0;
	}else if(strcmp(test_name, "allreduce") == 0)
	{
		return allreduce ;
	}else if(strcmp(test_name, "allgather") == 0)
	{
		return allgather ;
	}else if(strcmp(test_name, "bcast") == 0)
	{
		return bcast ;
	}else if(strcmp(test_name, "gather") == 0)
	{
		return gather ;
	}else if(strcmp(test_name, "barrier") == 0)
	{
		return barrier ;
	}else if(strcmp(test_name, "scatter") == 0)
	{
		return scatter ;
	}else if(strcmp(test_name, "sendRecvPaired") == 0)
	{
		return sendRecvPaired ;
	}else if(strcmp(test_name, "sendRecvRightNeighbour" ) == 0)
	{
		return sendRecvRightNeighbour ;
	}else if(strcmp(test_name, "sendToRoot") == 0)
	{
		return sendToRoot ;
	}else if(strcmp(test_name, "sendRecvToRoot") == 0)
	{
		return sendRecvToRoot ;
	}else {
		return NULL;
	}

	
	
}

