#ifndef MNB_UTILS
#define MNB_UTILS

#include<stdio.h>
#include<stdlib.h>
#include<mpi.h>
#include"LinkedList.h"
#include"mnb-environment.h"



#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif

//======================================================================
//defines to simplify debugging

#ifdef DEBUG
#define MNB_printf(s, ...); printf(s, ##__VA_ARGS__);
#else
#define MNB_printf(...);       ;
#endif


#ifdef DEBUG
#define MNB_println(line); { int printRank; MPI_Comm_rank(MPI_COMM_WORLD, &printRank); printf("%d:%s\n", printRank, line);}
#else
#define MNB_println(line);       ;
#endif

//======================================================================

//TODO zwei entfernen
//~ #define mpi_bench_perror_and_quit(s) { perror(s); exit(1); }

#define MNB_Abort(s) { perror(s); exit(1); }

//~ #define MNB_QUIT(s) { perror(s); exit(1); }


// all possigle types of MNB_Settings
enum MNB_SettingTypes { TYPE_NONE,
	TYPE_SIZE, 
	TYPE_TEST, 
	TYPE_NODE, 
	TYPE_VALUE,
	TYPE_INT,
	TYPE_STRING,
	TYPE_LIST,
	TYPE_HOST,
	TYPE_PROCESS,
	TYPE_MIN_REPEATS,
	TYPE_MAX_DURATION
	};



/**
 * a Setting containing a name , type and the setting value
 * 
 * 
 **/
typedef struct {
	void * data;
	int type;
	char * name;
} MNB_Setting;


/**
 * Allocates the requered memory and sets default values for a MNB_Setting
 **/
MNB_Setting * MNB_CreateSetting();

/**
 * Prints out the name, type and content of a setting
 * 
 **/
void MNB_PrintSetting(MNB_Setting * typed_setting);


/**
 * this structure contains all information for a test
 * 
 **/ 
typedef struct {
	char * name;
	char * test;
	struct MNB_List * sizes;
	struct MNB_List * nodeConf;  
	int minRepeats; //for both 0 is consideren an invalid value, and as a sign to use the global one.
	int maxDuration; //in seconds
} MNB_Test;

#define MNB_GetNodeRequerment(test) *((int*)(((MNB_Setting *)(((MNB_Test *)test)->nodeConf->first->data))->data))
#define MNB_GetRankRequerment(test) *((int*)(((MNB_Setting *)(((MNB_Test *)test)->nodeConf->first->next->data))->data))
#define MNB_GetResenderRank(node) *((int *)(((MNB_Node *)node)->ranks->first->data))


/**
 * the following four structures are used for communication of the ranks
 * each MNB_MessageBlock describes one test and the message bescribes all
 * tests of a round
 * 
 **/
typedef struct{
	int numberOfNodes;
	int numberOfRanks;
	int numberOfSizes;
	int sizeOfTestCollName;
	int sizeOfTestName;
	int sizeOfTestSource;
}MNB_MessageBlockHeader;

typedef struct{
	int * ranks;
	int * sizes;
	char * testCollName; //name of the test collection(makes it simpler to identify)
	char * testName;
	char * testSource;
	int minRepeats;
	int maxDuration;
}MNB_MessageBlockBody;

typedef struct {
	MNB_MessageBlockHeader * header;
	MNB_MessageBlockBody * body;
} MNB_MessageBlock;


typedef struct {
	int type;
	int numberOfBlocks;
	MNB_MessageBlock * messageBlocks;
} MNB_Message;

enum MNB_MESSAGE_TYPE { MNB_QUIT_MESSAGE, MNB_EXECUTE_MESSAGE};

/**
 * prints all information about a test
 * 
 **/
void MNB_PrintTest(MNB_Test * run_to_print);

/**
 * allocates the memory and sets the values for a MNB_Test
 **/
MNB_Test * MNB_CreateTest();


/**
 * return a copy of the tests
 * 
 **/
MNB_Test* MNB_CopyTest( MNB_Test * run_orig);



/**
 * is used by the tests to allocate memory
 * 
 **/
void * mallocWithCheck(long size);



/**
 * type of the testfunction pointer 
 * 
 * every test funktion has to accept the size of data the test schould 
 * be executed with and the communicator containing all ranks, who are
 * part of the test as parameters and return the time the test took
 * to execute
 * 
 **/
typedef double (*testFunc)(long size, MPI_Comm comm);




/**
 * creates a MPI_Group based on information provided in the messageBlock
 * (has to be called by all members of comm)
 * 
 **/
MPI_Comm * MNB_CreateCommFromMessageBlock(MNB_MessageBlock * messageBlock, MPI_Comm * comm);



/**
 * creates a MPI_Comm containing all ranks except the scheduler
 * (and the ranks on scheduler node, if we have more then one node)
 * 
 **/
MPI_Comm * MNB_CreateTesterComm(MNB_Environment * env);

int MNB_SizeStringToInt(char * size);


/**
 * rank 0 prints an welcome message for the user with some informatio about the environment
 * 
 **/
void MNB_PrintWelcome(MNB_Environment * env);


#endif
